import boto3
from contextlib import ExitStack
from dfmock import DFMock
import logging
import moto
import os
import pandas as pd
import numpy as np
from pandas.util.testing import assert_frame_equal
import pyarrow as pa
import pyarrow.parquet as pq
import random
from string import ascii_lowercase
import tempfile
from typing import Dict

logger = logging.getLogger(__name__)
logger.warning(
    "s3parq.testing_helper is made for internal use and may change unpredictably")


"""
Assertion-based helper functions:
- df_equal_by_set
- sorted_dfs_equal_by_pandas_testing

Setup-based helper functions:
- setup_files_list
- setup_grouped_dataframe
- setup_partitioned_parquet
- setup_random_string
"""


# START ASSERTION BASED


def df_equal_by_set(df1: pd.DataFrame, df2: pd.DataFrame, cols: iter) -> bool:
    """ This checks over whether 2 dataframes are equal by comparing zips of
    the given columns

    Args:
        df1 (pd.DataFrame): The first dataframe to compare against
        df2 (pd.DataFrame): The second dataframe to compare against
        cols (iter): The column to compare on the dataframes

    Returns:
        bool of whether the dataframe columns are equal
    """
    set1, set2 = list(), list()
    for col in cols:
        set1.append(df1[col])
        set2.append(df2[col])
    zipped1 = set(zip(*set1))
    zipped2 = set(zip(*set2))

    return (zipped1 == zipped2)


def sorted_dfs_equal_by_pandas_testing(df1: pd.DataFrame, df2: pd.DataFrame) -> None:
    """ Wraps the pandas.util.testing assert_frame_equal with the boilerplate
    of fixing any irregularities from the publish/fetch process - this
    includes index and sorting, as those will almost never match
    Args:
        df1 (pd.DataFrame): The first dataframe to compare
        df2 (pd.DataFrame): The second dataframe to compare
    Raises:
        AttributeError is the two dataframes are not equivalent
    """
    # make the column order match
    df2 = df2[df1.columns]
    # make them the same format, remerging partitions changes it all around
    df2 = df2.sort_values(
        by=df1.columns.tolist()).reset_index(drop=True)
    df1 = df1.sort_values(
        by=df1.columns.tolist()).reset_index(drop=True)
    # The setup on this part is a pain but it's the most specific check
    assert_frame_equal(df2, df1)


# START SETUP BASED


def setup_files_list(count: int = 100, bucket: str = None, key: str = None, s3_client=None):
    """ Creates temporary files and publishes them to the mocked S3 to test fetches,
    will fill in unsupplied parameters with random values or defaults

    Args:
        count (int, Optional): How many files to create, defaults to 100
        bucket (str, Optional): bucket to create to put parquet files in,
            will be a random string if not supplied
        key (str, Optional): S3 key to put parquet files in, will be a random
            string if not supplied
        partition_data_types (Dict, Optional): partition columns and datatypes,
            will be the default if not supplied

    Returns:
        A dict of retrieval informations (bucket, key, files)
    """
    if not key:
        key = setup_random_string()
    if not bucket:
        bucket = setup_random_string()
    temp_file_names = []
    s3_client.create_bucket(Bucket=bucket)

    with tempfile.TemporaryDirectory() as tmp_dir:
        for x in range(count):
            with tempfile.NamedTemporaryFile(dir=tmp_dir) as t:
                head, tail = os.path.split(t.name)
                temp_file_names.append(str(tail))
                with open(t.name, 'rb') as data:
                    s3_client.upload_fileobj(
                        data, Bucket=bucket, Key=(key + "/" + tail + ".parquet"))

    retrieval_ops = {
        "bucket": bucket,
        "key": key,
        "files": temp_file_names
    }

    return retrieval_ops


def setup_grouped_dataframe(count: int = 100, columns: Dict = None):
    """ Creates mock dataframe using dfmock

    Args:
        count (int): Row length to generate on the dataframe
        columns (Dict): dictionary of columns and types, following dfmock guides

    Returns:
        A freshly mocked dataframe
    """
    df = DFMock()
    df.count = count
    if not columns:
        columns = {"string_col": {"option_count": 3, "option_type": "string"},
                   "int_col": {"option_count": 3, "option_type": "int"},
                   "float_col": {"option_count": 3, "option_type": "float"},
                   "bool_col": {"option_count": 3, "option_type": "bool"},
                   "datetime_col": {"option_count": 3, "option_type": "datetime"},
                   "text_col": "string",
                   "metrics": "int"
                   }
    df.columns = columns
    df.generate_dataframe()
    return df.dataframe


def setup_partitioned_parquet(
    dataframe: pd.DataFrame = None,
    bucket: str = None,
    key: str = None,
    partition_data_types: Dict = None,
    s3_client=None
):
    """ Creates temporary files and publishes them to the mocked S3 to test fetches,
    will fill in unsupplied parameters with random values or defaults

    Args:
        dataframe (pd.DataFrame): dataframe to split into parquet files
        bucket (str, Optional): bucket to create to put parquet files in,
            will be a random string if not supplied
        key (str, Optional): S3 key to put parquet files in, will be a random
            string if not supplied
        partition_data_types (Dict, Optional): partition columns and datatypes,
            will be the default if not supplied
        s3_client (boto3 S3 client, Optional): The started S3 client that boto
            uses - NOTE: this should be made under a moto S3 mock!
            If it is not provided, a session is crafted under moto.mock_s3

    Returns:
        A tuple of the bucket and the published parquet file paths
    """
    if dataframe is None:
        dataframe = setup_grouped_dataframe()
    if not key:
        key = setup_random_string()
    if not bucket:
        bucket = setup_random_string()
    if partition_data_types is None:
        partition_data_types = {
            "string_col": "string",
            "int_col": "integer",
            "float_col": "float",
            "bool_col": "boolean",
            "datetime_col": "datetime"
        }

    with ExitStack() as stack:
        tmp_dir = stack.enter_context(tempfile.TemporaryDirectory())

        s3_client.create_bucket(Bucket=bucket)

        # generate the local parquet tree
        table = pa.Table.from_pandas(dataframe)
        pq.write_to_dataset(table,
                            root_path=str(tmp_dir),
                            partition_cols=list(
                                partition_data_types.keys())
                            )

        parquet_paths = []

        # traverse the local parquet tree
        extra_args = {'partition_data_types': str(
            partition_data_types)}
        for subdir, dirs, files in os.walk(str(tmp_dir)):
            for file in files:
                full_path = os.path.join(subdir, file)
                full_key = '/'.join([key,
                                     subdir.replace(f"{tmp_dir}/", ''), file])
                with open(full_path, 'rb') as data:
                    s3_client.upload_fileobj(data, Bucket=bucket, Key=full_key, ExtraArgs={
                        "Metadata": extra_args})
                    parquet_paths.append(full_key)

    return bucket, parquet_paths


def setup_nons3parq_parquet(
    dataframe: pd.DataFrame = None,
    bucket: str = None,
    key: str = None,
    s3_client=None
):
    """ Creates temporary files and publishes them to the mocked S3 to test fetches,
    will fill in unsupplied parameters with random values or defaults

    Args:
        dataframe (pd.DataFrame): dataframe to split into parquet files
        bucket (str, Optional): bucket to create to put parquet files in,
            will be a random string if not supplied
        key (str, Optional): S3 key to put parquet files in, will be a random
            string if not supplied
        s3_client (boto3 S3 client, Optional): The started S3 client that boto
            uses - NOTE: this should be made under a moto S3 mock!
            If it is not provided, a session is crafted under moto.mock_s3

    Returns:
        A tuple of the bucket and the published parquet file paths
    """
    if dataframe is None:
        dataframe = setup_grouped_dataframe()
    if not key:
        key = setup_random_string()
    if not bucket:
        bucket = setup_random_string()

    with ExitStack() as stack:
        tmp_dir = stack.enter_context(tempfile.TemporaryDirectory())
        if not s3_client:
            stack.enter_context(moto.mock_s3())
            s3_client = boto3.client('s3')

        s3_client.create_bucket(Bucket=bucket)

        # generate the local parquet tree
        table = pa.Table.from_pandas(dataframe)
        pq.write_to_dataset(table, root_path=str(tmp_dir), partition_cols=[])

        parquet_paths = []

        # traverse the local parquet tree
        for subdir, dirs, files in os.walk(str(tmp_dir)):
            for file in files:
                full_path = os.path.join(subdir, file)
                full_key = '/'.join([key,
                                     subdir.replace(f"{tmp_dir}/", ''), file])
                with open(full_path, 'rb') as data:
                    s3_client.upload_fileobj(data, Bucket=bucket, Key=full_key)
                    parquet_paths.append(full_key)

    return bucket, parquet_paths


def setup_random_string(min_len: int = 0, max_len: int = 10):
    """ Create a random string of either given min_lento max_len or default 0 to 10 """
    return ''.join([random.choice(ascii_lowercase) for x in range(min_len, max_len)])


def setup_custom_redshift_columns_and_dataframe():
    """ Create a custom_redshift_columns dictionary that contains redshift column definitions and corresponding mock dataframe """
    sample_data = {'colA': ["A", "B", "C"], 'colB': [4, 5, 6], 'colC': [4.12, 5.22, 6.57], 'colD': [
        4.1289, 5.22, 6.577], 'colE': ["test1", "test2", "test3"], 'colF': [True, False, True]}
    dataframe = pd.DataFrame(data=sample_data)

    custom_redshift_columns = {"colA": "VARCHAR(1000)",
                               "colB": "BIGINT",
                               "colC": "REAL",
                               "colD": "DECIMAL(5,4)",
                               "colE": "VARCHAR",
                               "colF": "BOOLEAN"}

    return (dataframe, custom_redshift_columns)


def setup_custom_redshift_columns_and_dataframe_with_null():
    """ Create a custom_redshift_columns dictionary that contains redshift column definitions and corresponding mock dataframe """
    sample_data = {'colA': [1, 2, np.NaN], 'colB': ['DDD', None, 'FFF'],
                   'colC': [pd.Timestamp('20131213 11:59:59.999999999'), None, pd.Timestamp('20131213 11:59:59.999999999')], 'colE': [7.5, 3.4, np.NaN]}

    dataframe = pd.DataFrame(data=sample_data)

    custom_redshift_columns = {"colA": "INTEGER",
                               "colB": "VARCHAR",
                               "colC": "TIMESTAMP",
                               "colE": "DECIMAL(7,2)"
                               }

    return (dataframe, custom_redshift_columns)



import boto3
import s3fs
import re
import sys
import logging
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
from typing import List
from s3parq.session_helper import SessionHelper
from s3parq import publish_redshift
from sqlalchemy import Column, Integer, String

logger = logging.getLogger(__name__)


def check_empty_dataframe(dataframe: pd.DataFrame) -> None:
    """ Make sure that the dataframe being published is not empty

    Args:
        dataframe (pd.DataFrame): Dataframe to verify

    Returns:
        None

    Raises:
        ValueError: If dataframe is empty
    """
    if dataframe.empty:
        no_data_message = "Empty dataframes cannot be published."
        raise ValueError(no_data_message)


def check_dataframe_for_timedelta(dataframe: pd.DataFrame) -> None:
    """ Pyarrow can't do timedelta, this verifies one is not present

    Args:
        dataframe (pd.DataFrame): Dataframe to verify

    Returns:
        None

    Raises:
        NotImplementedError: If the dataframe contains a timedelta dtype
    """
    dtypes = set(dataframe.columns)
    for dtype in dtypes:
        if 'timedelta' in dtype:
            timedelta_message = "Pyarrow does not support parquet conversion of timedelta columns at this time."
            logger.debug(timedelta_message)
            raise NotImplementedError(timedelta_message)


def _check_partition_compatibility(partition: str) -> bool:
    """ Make sure each partition value is hive-allowed.

    Args:
        partition (str): Partition to verify

    Returns:
        bool of whether the partition is not a hive reserved one
    """
    reserved = "ALL ALTER AND ARRAY AS AUTHORIZATION BETWEEN BIGINT BINARY BOOLEAN BOTH BY CASE CAST CHAR COLUMN CONF CREATE CROSS CUBE CURRENT CURRENT_DATE CURRENT_TIMESTAMP CURSOR DATABASE DATE DECIMAL DELETE DESCRIBE DISTINCT DOUBLE DROP ELSE END EXCHANGE EXISTS EXTENDED EXTERNAL FALSE FETCH FLOAT FOLLOWING FOR FROM FULL FUNCTION GRANT GROUP GROUPING HAVING IF IMPORT IN INNER INSERT INT INTERSECT INTERVAL INTO IS JOIN LATERAL LEFT LESS LIKE LOCAL MACRO MAP MORE NONE NOT NULL OF ON OR ORDER OUT OUTER OVER PARTIALSCAN PARTITION PERCENT PRECEDING PRESERVE PROCEDURE RANGE READS REDUCE REVOKE RIGHT ROLLUP ROW ROWS SELECT SET SMALLINT TABLE TABLESAMPLE THEN TIMESTAMP TO TRANSFORM TRIGGER TRUE TRUNCATE UNBOUNDED UNION UNIQUEJOIN UPDATE USER USING UTC_TMESTAMP VALUES VARCHAR WHEN WHERE WINDOW WITH COMMIT ONLY REGEXP RLIKE ROLLBACK START CACHE CONSTRAINT FOREIGN PRIMARY REFERENCES DAYOFWEEK EXTRACT FLOOR INTEGER PRECISION VIEWS TIME NUMERIC SYNC".split()
    return not (partition.upper() in reserved)


def check_partitions(partitions: iter, dataframe: pd.DataFrame) -> None:
    """ Make sure all given partitions are in the dataframe and HIVE-allowed

    Args:
        partitions (iter): Partitions to verify
        dataframe (pd.DataFrame): Dataframe the partitions should be in

    Returns:
        None

    Raises:
        ValueError: If the partition is not present in the dataframe's columns
        ValueError: If the partition is a HIVE reserved word
    """
    logger.debug("Checking partition args...")
    for partition in partitions:
        if partition not in dataframe.columns.tolist():
            partition_message = f"Cannot set {partition} as a partition; this is not a valid column header for the supplied dataframe."
            logger.critical(partition_message)
            raise ValueError(partition_message)
        if not _check_partition_compatibility(partition):
            partition_message = f"{partition} is a reserved word in hive that cannot be used as a partition."
            logger.critical(partition_message)
            raise ValueError(partition_message)

    logger.debug("Done checking partitions.")


def validate_redshift_params(redshift_params: dict) -> dict:
    """ Make sure redshift params are the expected format with all needed values
    Will also lowercase table and schema names.

    Args:
        redshift_params (dict):
            Dictionary for Spectrum, should be provided in the following format
            The params should be formatted as follows:

                - schema_name (str): Name of the Spectrum schema to publish to
                - table_name (str): Name of the table to write the dataset as
                - iam_role (str): Role to take while writing data to Spectrum
                - region (str): AWS region for Spectrum
                - cluster_id (str): Spectrum cluster id
                - host (str): Redshift Spectrum host name
                - port (str): Redshift Spectrum port to use
                - db_name (str): Redshift Spectrum database name to use
                - ec2_user (str): If on ec2, the user that should be used

    Returns:
        The given redshift_params, with table and schema names lowercase

    Raises:
        ValueError: If the length of redshift_params is less than expected
        ValueError: If redshift_params other than ec2_user have no value
        ValueError: If redshift_params is missing any of the above attributes
    """
    expected_params = ["schema_name", "table_name", "iam_role",
                       "region", "cluster_id", "host", "port", "db_name", "ec2_user"]
    logger.debug("Checking redshift params are correctly formatted")
    if len(redshift_params) != len(expected_params):
        params_length_message = f"Expected parameters: {len(expected_params)}. Received: {len(redshift_params)}"
        raise ValueError(params_length_message)
    for key, item in redshift_params.items():
        if not item and key != "ec2_user":
            params_type_message = f"No value assigned for param {key}."
            raise ValueError(params_type_message)
    for param in expected_params:
        if param not in redshift_params.keys():
            params_key_message = f"Error: Required parameter {param} not found in passed redshift_params."
            raise KeyError(params_key_message)

     # Ensure redshift table name is lowercase
    if redshift_params['table_name'] != redshift_params['table_name'].lower():
        logger.warning(
            f"Table name {redshift_params['table_name']} contains uppercase letters. Converting to lowercase...")
        redshift_params['table_name'] = redshift_params['table_name'].lower()

    # Ensure redshift schema name is lowercase
    if redshift_params['schema_name'] != redshift_params['schema_name'].lower():
        logger.warning(
            f"Schema name {redshift_params['schema_name']} contains uppercase letters. Converting to lowercase...")
        redshift_params['schema_name'] = redshift_params['schema_name'].lower()

    logger.debug('Done validating redshift params')
    return redshift_params


def s3_url(bucket: str, key: str) -> str:
    """ Turn a bucket + key into a S3 url """
    return '/'.join(["s3:/", bucket, key])


def _gen_parquet_to_s3(bucket: str, key: str, dataframe: pd.DataFrame,
                       partitions: list, custom_redshift_columns: dict = None) -> None:
    """ Converts the dataframe into a PyArrow table then writes it into S3 as
    parquet. Partitions are done with PyArrow parquet's write.
    NOTE: Timestamps are coerced to ms , this is due to a constraint of parquet
        with pandas's datetime64

    Args:
        bucket (str): S3 bucket to publish to
        key (str): S3 key to the root of where the dataset should be published
        dataframe (pd.DataFrame): Dataframe that's being published
        partitions (list): List of partition columns
        custom_redshift_columns (dict): 
            This dictionary contains custom column data type definitions for redshift.
            The params should be formatted as follows:
                - column name (str)
                - data type (str)

    Returns:
        None
    """
    logger.info("Writing to S3...")

    try:
        (schema, frame) = _parquet_schema(dataframe,
                                          custom_redshift_columns=custom_redshift_columns)
        table = pa.Table.from_pandas(
            df=frame, schema=schema, preserve_index=False)
    except pa.lib.ArrowTypeError:
        logger.warning(
            "Dataframe conversion to pyarrow table failed, checking object columns for mixed types")
        dataframe_dtypes = dataframe.dtypes.to_dict()
        object_columns = [
            # Skip conversion to strings if the column contains decimal objects. Otherwise, decimal objects will be improperly converted.
            col for col, col_type in dataframe_dtypes.items() if col_type == "object" and "[Decimal(" not in str(dataframe[col].values)[:9]]
        for object_col in object_columns:
            if not dataframe[object_col].apply(isinstance, args=[str]).all():
                logger.warning(
                    f"Dataframe column : {object_col} : in this chunk is type object but contains non-strings, converting to all-string column")
                dataframe[object_col] = dataframe[object_col].astype(str)

        logger.info("Retrying conversion to pyarrow table")
        (schema, frame) = _parquet_schema(dataframe,
                                          custom_redshift_columns=custom_redshift_columns)
        table = pa.Table.from_pandas(
            df=frame, schema=schema, preserve_index=False)

    uri = s3_url(bucket, key)
    logger.debug(f"Writing to s3 location: {uri}...")
    pq.write_to_dataset(table, compression="snappy", root_path=uri,
                        partition_cols=partitions, filesystem=s3fs.S3FileSystem(), coerce_timestamps='ms', allow_truncated_timestamps=True)
    logger.debug("Done writing to location.")


def _assign_partition_meta(bucket: str, key: str, dataframe: pd.DataFrame, partitions: List['str'], session_helper: SessionHelper, redshift_params=None, custom_redshift_columns: dict = None) -> List[str]:
    """ Assigns the dataset partition meta to all object keys in the dataset.
    Keys are found by listing all files under the given key and then filtering
    to only those that end in '.parquet' and then further filtering to those
    that do not already have partition_data_types metadata.

    Args:
        bucket (str): S3 bucket to publish to
        key (str): S3 key to the root of where the dataset is published
        dataframe (pd.DataFrame): Dataframe that has been published
        partitions (list): List of partition columns
        session_helper (SessionHelper): Current session, if not using Spectrum this should be None
        redshift_params (dict, Optional):
            Dictionary for Spectrum, should be in the following format
            The params should be formatted as follows:

                - schema_name (str): Name of the Spectrum schema to publish to
                - table_name (str): Name of the table to write the dataset as
                - iam_role (str): Role to take while writing data to Spectrum
                - region (str): AWS region for Spectrum
                - cluster_id (str): Spectrum cluster id
                - host (str): Redshift Spectrum host name
                - port (str): Redshift Spectrum port to use
                - db_name (str): Redshift Spectrum database name to use
                - ec2_user (str): If on ec2, the user that should be used
        custom_redshift_columns (dict, Optional): 
            This dictionary contains custom column data type definitions for redshift.
            The params should be formatted as follows:
                - column name (str)
                - data type (str)

    Returns:
        A str list of object keys of the objects that got metadata added
    """
    s3_client = boto3.client('s3')
    all_files_without_meta = []
    paginator = s3_client.get_paginator('list_objects')
    page_iterator = paginator.paginate(Bucket=bucket, Prefix=key)
    for page in page_iterator:
        for obj in page['Contents']:
            if obj['Key'].endswith(".parquet"):
                head_obj = s3_client.head_object(Bucket=bucket, Key=obj['Key'])
                if not 'partition_data_types' in head_obj['Metadata']:
                    all_files_without_meta.append(obj['Key'])
                    if redshift_params and partitions:
                        sql_command = publish_redshift.create_partitions(
                            bucket, redshift_params['schema_name'], redshift_params['table_name'], obj['Key'], session_helper)

    for obj in all_files_without_meta:
        logger.debug(f"Appending metadata to file {obj}..")
        s3_client.copy_object(Bucket=bucket, CopySource={'Bucket': bucket, 'Key': obj}, Key=obj,
                              Metadata={'partition_data_types': str(
                                  _parse_dataframe_col_types(
                                      dataframe=dataframe, partitions=partitions, custom_redshift_columns=custom_redshift_columns)
                              )}, MetadataDirective='REPLACE')
        logger.debug("Done appending metadata.")
    return all_files_without_meta


def _get_dataframe_datatypes(dataframe: pd.DataFrame, partitions=[], use_parts=False) -> dict:
    """ Gets the dtypes of the columns in the dataframe

    Args:
        dataframe (pd.DataFrame): Dataframe that has been published
        partitions (list, Optional): List of partition columns
        use_parts (bool, Optional): bool to determine if only partition datatypes
            should be returned

    Returns:
        Dictionary of the column names as keys and dtypes as values,
            if use_parts is true then only limited to partition columns
    """
    types = dict()
    if use_parts:
        columns = partitions
    else:
        columns = dataframe.drop(labels=partitions, axis="columns").columns
    for col in columns:
        type_string = dataframe[col].dtype.name
        types[col] = type_string
    return types


def _parquet_schema(dataframe: pd.DataFrame, custom_redshift_columns: dict = None):
    """ Translates pandas dtypes to PyArrow types and creates a Schema from them

    Args:
        dataframe (pd.DataFrame): Dataframe to pull the schema of
        custom_redshift_columns (dict, Optional): 
            This dictionary contains custom column data type definitions for redshift.
            The params should be formatted as follows:
                - column name (str)
                - data type (str)

    Returns:
        PyArrow Schema of the given dataframe
        Potentially modified Dataframe
    """
    fields = []
    for col, dtype in dataframe.dtypes.items():
        dtype = dtype.name
        if dtype == 'object':
            if custom_redshift_columns:
                # Detect if the Pandas object column contains Python decimal objects.
                if "[Decimal(" in str(dataframe[col].values)[:9]:
                    # If Python decimal objects are present, parse out the precision and scale
                    # from the custom_redshift_columns dictionary to use when converting
                    # to PyArrow's decimal128 data type.
                    s = custom_redshift_columns[col]
                    precision = int(
                        s[s.find('DECIMAL(') + len('DECIMAL('):s.rfind(',')].strip())
                    scale = int(s[s.find(',') + len(','):s.rfind(')')].strip())
                    pa_type = pa.decimal128(precision=precision, scale=scale)
                else:
                    pa_type = pa.string()
            else:
                pa_type = pa.string()
        elif dtype.startswith('int32'):
            pa_type = pa.int32()
        elif dtype.startswith('int64'):
            pa_type = pa.int64()
        elif dtype.startswith('int8'):
            pa_type = pa.int8()
        elif dtype.startswith('Int32'):
            dataframe = dataframe.astype({col: 'object'})
            pa_type = pa.int32()
        elif dtype.startswith('Int64'):
            dataframe = dataframe.astype({col: 'object'})
            pa_type = pa.int64()
        elif dtype.startswith('float32'):
            pa_type = pa.float32()
        elif dtype.startswith('float64'):
            pa_type = pa.float64()
        elif dtype.startswith('float16'):
            pa_type = pa.float16()
        elif dtype.startswith('datetime'):
            pa_type = pa.timestamp('ns')
        elif dtype.startswith('date'):
            pa_type = pa.date64()
        elif dtype.startswith('category'):
            pa_type = pa.string()
        elif dtype == 'bool':
            pa_type = pa.bool_()
        else:
            raise NotImplementedError(
                f"Error: {dtype} is not a datatype which can be mapped to Parquet using s3parq.")
        fields.append(pa.field(col, pa_type))

    return (pa.schema(fields=fields), dataframe)


def _parse_dataframe_col_types(dataframe: pd.DataFrame, partitions: list, custom_redshift_columns: dict = None) -> dict:
    """ Determines the metadata of the partition columns based on dataframe dtypes

    Args:
        dataframe (pd.DataFrame): Dataframe to parse the types of
        partitions (list): Partitions in the dataframe to get the type of
        custom_redshift_columns (dict, Optional):
            This dictionary contains custom column data type definitions for redshift.
            The params should be formatted as follows:
                - column name (str)
                - data type (str)

    Returns:
        Dictionary of column names as keys with their datatypes (string form) as values
    """
    logger.debug("Determining write metadata for publish...")
    dataframe = dataframe[partitions]
    dtypes = {}
    for col, dtype in dataframe.dtypes.items():
        dtype = dtype.name
        if dtype == 'object':
            dtypes[col] = 'string'
        elif dtype.startswith('int'):
            dtypes[col] = 'integer'
        elif dtype.startswith('float'):
            dtypes[col] = 'float'
        elif dtype.startswith('date'):
            dtypes[col] = 'datetime'
        elif dtype.startswith('category'):
            dtypes[col] = 'category'
        elif dtype == 'bool':
            dtypes[col] = 'boolean'

        if custom_redshift_columns:
            if 'DECIMAL' in custom_redshift_columns[col]:
                dtypes[col] = 'decimal'

    logger.debug(f"Done. Metadata determined as {dtypes}")
    return dtypes


def _sized_dataframes(dataframe: pd.DataFrame) -> tuple:
    """ Determines optimal chunks to publish the dataframe in. In smaller dataframes
    this may be the whole dataframe.
    This is determined by the hard values of 60MB being ideal for Spectrum, and
    assumed compression ratio for dataframes to parquet being ~4; the bytes per
    row are then estimated and then chunks are determined.

    Args:
        dataframe (pd.DataFrame): Dataframe to size out

    Yields:
        A dictionary containing the upper and lower index of the dataframe to chunk
    """

    def log_size_estimate(num_bytes):
        if num_bytes == 0:
            return "0MB"
        elif round(num_bytes / float(1 << 20), 2) == 0.0:
            return "<0.1MB"
        else:
            return str(round(num_bytes / float(1 << 20), 2)) + "MB"

    # get first row size
    row_size_est = sys.getsizeof(dataframe.head(1))
    # get number of rows
    num_rows = int(dataframe.shape[0])
    frame_size_est = row_size_est * num_rows
    # at scale dataframes seem to compress around 3.5-4.5 times as parquet.
    # TODO: should build a real regression to calc this!
    compression_ratio = 4
    # 60MB compressed is ideal for Spectrum
    ideal_size = compression_ratio * (60 * float(1 << 20))
    # short circut if < ideal size
    batch_log_message = f"""row size estimate: {row_size_est} bytes.
number of rows: {num_rows} rows
frame size estimate: {frame_size_est} bytes
compression ratio: {compression_ratio}:1
ideal size: {ideal_size} bytes
"""
    logger.debug(batch_log_message)
    if ideal_size > frame_size_est:
        yield {'lower': 0, 'upper': len(dataframe)}
        return

    # math the number of estimated partitions
    sized_frames = []
    num_partitions = int(row_size_est * num_rows / ideal_size)
    rows_per_partition = int(num_rows / num_partitions)
    # for each partition do the thing
    logger.info(
        f"Sized out {len(range(0, num_rows, rows_per_partition))} dataframes.")
    for index, lower in enumerate(range(0, num_rows, rows_per_partition)):
        if index == num_partitions:
            upper = num_rows
        else:
            upper = lower + rows_per_partition

        yield {'lower': lower, 'upper': upper}


def publish(bucket: str, key: str, partitions: List[str], dataframe: pd.DataFrame, redshift_params: dict = None) -> List[str]:
    """ Dataframe to S3 Parquet Publisher
    This function handles the portion of work that will see a dataframe converted
    to parquet and then published to the given S3 location.
    It supports partitions and will preserve the datatypes on s3parq fetch.
    It also has the option to automatically publish up to Redshift Spectrum for
    the newly published parquet files.

    Args:
        bucket (str): S3 Bucket name
        key (str): S3 key to lead to the desired dataset
        partitions (List[str]): List of columns that should be partitioned on
        dataframe (pd.DataFrame): Dataframe to be published
        redshift_params (dict, Optional):
            This dictionary should be provided in the following format in order
            for data to be published to Spectrum. Leave out entirely to avoid
            publishing to Spectrum.
            The params should be formatted as follows:

                - schema_name (str): Name of the Spectrum schema to publish to
                - table_name (str): Name of the table to write the dataset as
                - iam_role (str): Role to take while writing data to Spectrum
                - region (str): AWS region for Spectrum
                - cluster_id (str): Spectrum cluster id
                - host (str): Redshift Spectrum host name
                - port (str): Redshift Spectrum port to use
                - db_name (str): Redshift Spectrum database name to use
                - ec2_user (str): If on ec2, the user that should be used

    Returns:
        A str list of all the newly published object keys
    """
    session_helper = None

    if redshift_params:
        if "index" in dataframe.columns:
            raise ValueError(
                "'index' is a reserved keyword in Redshift. Please remove or rename your DataFrame's 'index' column.")

        logger.debug(
            "Found redshift parameters. Checking validity of params...")
        redshift_params = validate_redshift_params(redshift_params)
        logger.debug("Redshift parameters valid. Opening Session helper.")
        session_helper = SessionHelper(
            region=redshift_params['region'],
            cluster_id=redshift_params['cluster_id'],
            host=redshift_params['host'],
            port=redshift_params['port'],
            db_name=redshift_params['db_name'],
            ec2_user=redshift_params['ec2_user']
        )

        session_helper.configure_session_helper()
        publish_redshift.create_schema(
            redshift_params['schema_name'], redshift_params['db_name'], redshift_params['iam_role'], session_helper)
        logger.debug(
            f"Schema {redshift_params['schema_name']} created. Creating table {redshift_params['table_name']}...")

        df_types = _get_dataframe_datatypes(dataframe, partitions)
        partition_types = _get_dataframe_datatypes(dataframe, partitions, True)
        publish_redshift.create_table(
            redshift_params['table_name'], redshift_params['schema_name'], df_types, partition_types, s3_url(bucket, key), session_helper)
        logger.debug(f"Table {redshift_params['table_name']} created.")

    logger.debug("Checking publish params...")
    check_empty_dataframe(dataframe)
    check_dataframe_for_timedelta(dataframe)
    check_partitions(partitions, dataframe)
    logger.debug("Publish params valid.")
    logger.debug("Begin writing to S3..")

    files = []
    for frame_params in _sized_dataframes(dataframe):
        logger.info(
            f"Publishing dataframe chunk : {frame_params['lower']} to {frame_params['upper']}")
        frame = pd.DataFrame(
            dataframe[frame_params['lower']:frame_params['upper']])
        _gen_parquet_to_s3(bucket=bucket,
                           key=key,
                           dataframe=frame,
                           partitions=partitions)

        published_files = _assign_partition_meta(bucket=bucket,
                                                 key=key,
                                                 dataframe=frame,
                                                 partitions=partitions,
                                                 session_helper=session_helper,
                                                 redshift_params=redshift_params)
        files = files + published_files

    logger.info("Done writing to S3.")

    return files


def custom_publish(bucket: str, key: str, partitions: List[str], dataframe: pd.DataFrame, custom_redshift_columns: dict, redshift_params: dict = None) -> List[str]:
    """ Dataframe to S3 Parquet Publisher with a CUSTOM redshift column definition.
    Custom publish allows custom defined redshift column definitions to be used and 
    enables support for Redshift's decimal data type. 
    This function handles the portion of work that will see a dataframe converted
    to parquet and then published to the given S3 location.
    It supports partitions and will use the custom redshift columns defined in the
    custom_redshift_columns dictionary when creating the table schema for the parquet file. 
    View the Custom Publishes section of s3parq's readme file for more guidance on formatting
    the custom_redshift_columns dictionary. It also has the option to automatically publish up 
    to Redshift Spectrum for the newly published parquet files. 

    Args:
        bucket (str): S3 Bucket name
        key (str): S3 key to lead to the desired dataset
        partitions (List[str]): List of columns that should be partitioned on
        dataframe (pd.DataFrame): Dataframe to be published
        custom_redshift_columns (dict): 
            This dictionary contains custom column data type definitions for redshift.
            The params should be formatted as follows:
                - column name (str)
                - data type (str)
        redshift_params (dict, Optional):
            This dictionary should be provided in the following format in order
            for data to be published to Spectrum. Leave out entirely to avoid
            publishing to Spectrum.
            The params should be formatted as follows:
                - schema_name (str): Name of the Spectrum schema to publish to
                - table_name (str): Name of the table to write the dataset as
                - iam_role (str): Role to take while writing data to Spectrum
                - region (str): AWS region for Spectrum
                - cluster_id (str): Spectrum cluster id
                - host (str): Redshift Spectrum host name
                - port (str): Redshift Spectrum port to use
                - db_name (str): Redshift Spectrum database name to use
                - ec2_user (str): If on ec2, the user that should be used

    Returns:
        A str list of all the newly published object keys
    """
    logger.debug(
        "Running custom publish...")

    session_helper = None

    if redshift_params:
        if "index" in dataframe.columns:
            raise ValueError(
                "'index' is a reserved keyword in Redshift. Please remove or rename your DataFrame's 'index' column.")

        logger.debug(
            "Found redshift parameters. Checking validity of params...")
        redshift_params = validate_redshift_params(redshift_params)
        logger.debug("Redshift parameters valid. Opening Session helper.")
        session_helper = SessionHelper(
            region=redshift_params['region'],
            cluster_id=redshift_params['cluster_id'],
            host=redshift_params['host'],
            port=redshift_params['port'],
            db_name=redshift_params['db_name'],
            ec2_user=redshift_params['ec2_user']
        )

        session_helper.configure_session_helper()
        publish_redshift.create_schema(
            redshift_params['schema_name'], redshift_params['db_name'], redshift_params['iam_role'], session_helper)
        logger.debug(
            f"Schema {redshift_params['schema_name']} created. Creating table {redshift_params['table_name']}...")

        publish_redshift.create_custom_table(
            redshift_params['table_name'], redshift_params['schema_name'], partitions, s3_url(bucket, key), custom_redshift_columns, session_helper)
        logger.debug(f"Custom table {redshift_params['table_name']} created.")

    logger.debug("Checking publish params...")
    check_empty_dataframe(dataframe)
    check_dataframe_for_timedelta(dataframe)
    check_partitions(partitions, dataframe)
    logger.debug("Publish params valid.")
    logger.debug("Begin writing to S3..")

    files = []
    for frame_params in _sized_dataframes(dataframe):
        logger.info(
            f"Publishing dataframe chunk : {frame_params['lower']} to {frame_params['upper']}")
        frame = pd.DataFrame(
            dataframe[frame_params['lower']:frame_params['upper']])
        _gen_parquet_to_s3(bucket=bucket,
                           key=key,
                           dataframe=frame,
                           partitions=partitions,
                           custom_redshift_columns=custom_redshift_columns)

        published_files = _assign_partition_meta(bucket=bucket,
                                                 key=key,
                                                 dataframe=frame,
                                                 partitions=partitions,
                                                 session_helper=session_helper,
                                                 redshift_params=redshift_params,
                                                 custom_redshift_columns=custom_redshift_columns)
        files = files + published_files

    logger.info("Done writing to S3.")

    return files



from s3parq.session_helper import SessionHelper
from sqlalchemy import Column, Integer, String
import logging
import re
import pandas as pd

logger = logging.getLogger(__name__)


def _is_reserved_keyword(name: str) -> bool:
    """ Returns bool of whether the uppercased name is reserved in Redshift
    NOTE: This is backwards to expected by name (if _is_reserved_keyword , its fine)
    """
    reserved = "AES128 AES256 ALL ALLOWOVERWRITE ANALYSE ANALYZE AND ANY ARRAY AS ASC AUTHORIZATION BACKUP BETWEEN BINARY BLANKSASNULL BOTH BYTEDICT BZIP2 CASE CAST CHECK COLLATE COLUMN CONSTRAINT CREATE CREDENTIALS CROSS CURRENT_DATE CURRENT_TIME CURRENT_TIMESTAMP CURRENT_USER CURRENT_USER_ID DEFAULT DEFERRABLE DEFLATE DEFRAG DELTA DELTA32K DESC DISABLE DISTINCT DO ELSE EMPTYASNULL ENABLE ENCODE ENCRYPT ENCRYPTION END EXCEPT EXPLICIT FALSE FOR FOREIGN FREEZE FROM FULL GLOBALDICT256 GLOBALDICT64K GRANT GROUP GZIP HAVING IDENTITY IGNORE ILIKE IN INITIALLY INNER INTERSECT INTO IS ISNULL JOIN LANGUAGE LEADING LEFT LIKE LIMIT LOCALTIME LOCALTIMESTAMP LUN LUNS LZO LZOP MINUS MOSTLY13 MOSTLY32 MOSTLY8 NATURAL NEW NOT NOTNULL NULL NULLS OFF OFFLINE OFFSET OID OLD ON ONLY OPEN OR ORDER OUTER OVERLAPS PARALLEL PARTITION PERCENT PERMISSIONS PLACING PRIMARY RAW READRATIO RECOVER REFERENCES RESPECT REJECTLOG RESORT RESTORE RIGHT SELECT SESSION_USER SIMILAR SNAPSHOT SOME SYSDATE SYSTEM TABLE TAG TDES TEXT255 TEXT32K THEN TIMESTAMP TO TOP TRAILING TRUE TRUNCATECOLUMNS UNION UNIQUE USER USING VERBOSE WALLET WHEN WHERE WITH WITHOUT".split()
    return not(name.upper() in reserved)


def _validate_name(name: str) -> tuple:
    """ Returns tuple, of which the first index indicates if the name is valid,
    and the second is the error message of why it is invalid if so
    """
    if not _is_reserved_keyword(name):
        return tuple([False, f'name: {name} cannot be a reserved SQL keyword'])
    elif not bool(re.match(r"^[a-zA-Z0-9_]", name)):
        return tuple([False, f'name: {name} can only start with an alphanumeric or an underscore'])
    elif bool(re.search("([ '\"])", name)):
        return tuple([False, f'name: {name} cannot contain spaces or quotations'])
    elif len(name) < 1 or len(name) > 127:
        return tuple([False, f'name: {name} must be between 1 and 127 characters'])
    else:
        return tuple([True, None])


def _redshift_name_validator(*args) -> None:
    """ Passes a list of args through name validation

    Returns:
        None

    Raises:
        ValueError: Uses internal validate_name function and raises error based
            on failure reasons
    """
    for arg in args:
        response = _validate_name(arg)
        if not response[0]:
            raise ValueError(response[1])


def _get_partitions_for_spectrum(filename: str) -> [str]:
    '''
    Turns S3 filepath with partitions into list of only those partitions as strings
    Args:
        filename (str): entire filepath for a single file
    ----
    Returns:
        final_partitions (list of strings): these are the partitions for that file
    --------
    Example:
        Args:
            filename = 'some-path/to/data/zipcode=12345/birth_month=january/final_data_set.parquet
        ----
        Returns:
            final_partitions = ['zipcode=12345', 'birth_month=january']
    '''
    filepath = filename.split('/')
    final_partitions = [_dir for _dir in filepath if '=' in _dir]
    return final_partitions


def _format_partition_strings_for_sql(partitions: [str]) -> [str]:
    '''
    Formats a list of S3 partition strings for SQL statements
    Args:
        partitions ([str]): list of strings representing the partitions in S3
    ----
    Returns:
        formatted_partitions ([str]): list of the same partitions with quotes
        to use for SQL
    --------
    Example:
        Args:
            partitions = ['hamburger=abcd', 'hot_dog=1234']
        ----
        Returns:
            formatted_partitions = ["hamburger='abcd'", "hot_dog='1234'"]
    '''
    formatted_partitions = []
    for p in partitions:
        key, value = p.split('=')
        value = f"'{value}'"
        formatted_partitions.append(f'{key}={value}')
    return formatted_partitions


def _last_index_containing_substring(the_list: [str], substring: str) -> int:
    '''
    Returns index of last string that contains a substring within a list.  If there's no match, it returns
    the length of the list + 1 because we want the function that uses it "_get_partition_location"
    to not identify any partitions if none exists
    Args:
        the_list ([str]): list of strings, probably representing the partitions
        substring (str): string to look for in the_list
    ----
    Returns:
        i (int): The index in the list that contains the first instance of the substring
        returns -1 if no matches
    --------
    Example:
        Args: 
            the_list = ['path', 'to', 'data', 'hamburger=abcd', 'hot_dog=1234', 'the_file.parquet]
            substring = '='
        ----
        Returns:
            i = 3

    '''
    for s in reversed(the_list):
        if substring in s:
            return len(the_list) - the_list.index(s)
    return len(the_list) + 1


def _get_partition_location(filepath: str):
    '''
    Gets location of data in S3 for partitioning in S3.  You need to know the path to 
    the first partition in order to make a proper Spectrum partition w/ SQL-Redshift
    Args:
        filepath (str): path to a file in S3 that is partitioned
    Returns:
        final_path (str): path within S3 bucket that has the last partition of a filepath
    Example:
        Args: 
            filepath = 'path/to/data/hamburger=abcd/hot_dog=1234/abcd1234.parquet'
        Returns:
            final_path = 'path/to/data/hamburger=abcd/hot_dog=1234'
    '''
    separate_dirs = filepath.split('/')
    last_partition = _last_index_containing_substring(separate_dirs, "=")
    # I think this is more confusing than it has to be
    final_set = separate_dirs[:-last_partition + 1]
    final_path = '/'.join(final_set)
    if final_path == '':
        raise ValueError(f'No partitions in this filepath {filepath}')
    return final_path


def _datatype_mapper(columns: dict) -> dict:
    """ Takes a dict of column names and pandas dtypes and returns a redshift create table statement of column names/dtypes."""
    logger.debug("Determining write metadata for publish...")
    dtypes = {}
    sql_statement = ""
    for col, dtype in columns.items():
        if dtype == 'object':
            dtypes[col] = 'VARCHAR'
        elif dtype.startswith('int8'):
            dtypes[col] = 'INTEGER'
        elif dtype.startswith('int32'):
            dtypes[col] = 'INTEGER'
        elif dtype.startswith('int64'):
            dtypes[col] = 'BIGINT'
        elif dtype.startswith('float16'):
            dtypes[col] = 'REAL'
        elif dtype.startswith('float32'):
            dtypes[col] = 'REAL'
        elif dtype.startswith('float64'):
            dtypes[col] = 'FLOAT'
        elif dtype.startswith('date'):
            dtypes[col] = 'TIMESTAMP'
        elif dtype.startswith('category'):
            dtypes[col] = 'VARCHAR'
        elif dtype == 'bool':
            dtypes[col] = 'BOOLEAN'
        else:
            raise ValueError(
                f"Error: {dtype} is not a datatype which can be mapped to Redshift.")
        sql_statement += f'{col} {dtypes[col]}, '
    return f"({sql_statement[:-2]})"  # Slice off the last space and comma


def create_schema(schema_name: str, db_name: str, iam_role: str, session_helper: SessionHelper) -> None:
    """ Creates a schema in AWS redshift using a given iam_role

    Args:
        schema_name (str): Name of the schema to create in Redshift Spectrum
        db_name (str): (Existing) database that the schema should belong to
        iam_role (str): link to an existing AWS IAM Role with Redshift Spectrum 
            write permissions
        session_helper (str): Active and configured session_helper session to use
    """
    _redshift_name_validator(schema_name, db_name)
    with session_helper.db_session_scope() as scope:
        new_schema_query = f"CREATE EXTERNAL SCHEMA IF NOT EXISTS {schema_name} \
                FROM DATA CATALOG \
                database '{db_name}' \
                iam_role '{iam_role}';"

        logger.info(f'Running query to create schema: {new_schema_query}')
        scope.execute(new_schema_query)


def create_table(table_name: str, schema_name: str, columns: dict, partitions: dict, path: str, session_helper: SessionHelper) -> None:
    """ Creates a table in AWS redshift. The table will be named 
    schema_name.table_name and belong to the (existing) Redshift db db_name

    Args:
        table_name (str): name of created table
            NOTE: THIS WILL ERROR IF table_name ALREADY EXISTS IN REDSHIFT
        schema_name (str): name of schema in redshift; Schema must be external 
            and already exist!
        columns (dict): Dictionary with keys corresponding to column names and 
            values corresponding to pandas dtypes, excluding partition columns
        partitions (dict): Dict similar to columns, except ONLY with partition columns
        path (str): Path to published dataset in s3 (excluding partitions)
        session_helper (SessionHelper): Instance of Redshift s3parq.session_helper
    """
    _redshift_name_validator(table_name)
    redshift_columns = _datatype_mapper(columns)
    redshift_partitions = _datatype_mapper(partitions)
    with session_helper.db_session_scope() as scope:
        if_exists_query = f'SELECT EXISTS(SELECT schemaname, tablename FROM SVV_EXTERNAL_TABLES WHERE tablename=\'{table_name}\' AND schemaname=\'{schema_name}\');'
        table_exists = scope.execute(if_exists_query).first()[0]
        if table_exists:
            return

        if not partitions:
            new_schema_query = (
                f'CREATE EXTERNAL TABLE {schema_name}.{table_name} {redshift_columns} \
                STORED AS PARQUET \
                LOCATION \'{path}\';'
            )
        else:
            new_schema_query = (
                f'CREATE EXTERNAL TABLE {schema_name}.{table_name} {redshift_columns} \
                PARTITIONED BY {redshift_partitions} STORED AS PARQUET \
                LOCATION \'{path}\';'
            )
        logger.info(f'Running query to create table: {new_schema_query}')
        scope.execute(new_schema_query)

def create_custom_table(table_name: str, schema_name: str, partitions: dict, path: str, custom_redshift_columns: dict, session_helper: SessionHelper) -> None:
    """ Creates a table in AWS redshift. The table will be named 
    schema_name.table_name and belong to the (existing) Redshift db db_name.
    The created table will use the CUSTOM redshift column data types defined 
    in custom_redshift_columns.

    Args:
        table_name (str): name of created table
            NOTE: THIS WILL ERROR IF table_name ALREADY EXISTS IN REDSHIFT
        schema_name (str): name of schema in redshift; Schema must be external 
            and already exist!
        partitions (dict): Dict similar to columns, except ONLY with partition columns
        path (str): Path to published dataset in s3 (excluding partitions)
        custom_redshift_columns (dict): 
            This dictionary contains custom column data type definitions for redshift.
            The params should be formatted as follows:
                - column name (str)
                - data type (str)
        session_helper (SessionHelper): Instance of Redshift s3parq.session_helper
    """

    logger.info("Running create_custom_table...")

    _redshift_name_validator(table_name)

    logger.info("Generating create columns sql statement with custom redshift columns...")
    logger.info("Generating create partitions sql statement with custom redshift columns...")
    redshift_columns_sql = ""
    redshift_partitions_sql = ""
    for k, v in custom_redshift_columns.items():
        if k in partitions:
            redshift_partitions_sql += f'{k} {v}, '
        else:
            redshift_columns_sql += f'{k} {v}, '
    redshift_columns = f"({redshift_columns_sql[:-2]})"  # Slice off the last space and comma
    redshift_partitions = f"({redshift_partitions_sql[:-2]})"  # Slice off the last space and comma

    with session_helper.db_session_scope() as scope:
        if_exists_query = f'SELECT EXISTS(SELECT schemaname, tablename FROM SVV_EXTERNAL_TABLES WHERE tablename=\'{table_name}\' AND schemaname=\'{schema_name}\');'
        table_exists = scope.execute(if_exists_query).first()[0]
        if table_exists:
            return

        if not partitions:
            new_schema_query = (
                f'CREATE EXTERNAL TABLE {schema_name}.{table_name} {redshift_columns} \
                STORED AS PARQUET \
                LOCATION \'{path}\';'
            )
        else:
            new_schema_query = (
                f'CREATE EXTERNAL TABLE {schema_name}.{table_name} {redshift_columns} \
                PARTITIONED BY {redshift_partitions} STORED AS PARQUET \
                LOCATION \'{path}\';'
            )
        logger.info(f'Running query to create table: {new_schema_query}')
        scope.execute(new_schema_query)


def create_partitions(bucket: str, schema: str, table: str, filepath: str, session_helper: SessionHelper) -> None:
    ''' Executes the SQL that creates partitions on the given table for an 
    individual file

    Args:
        bucket (str): S3 bucket where data is stored
        schema (str): name of redshift schema (must already exist)
        table (str): name of table in schema.  Must have partitions scoped out in `CREATE TABLE ...`
        filepath (str): path to data in S3 that will be queryable by it's partitions
            NOTE: This is to the single parquet file, including the partitions
        session_helper (SessionHelper): a configured s3parq.session_helper.SessionHelper session

    Returns:
        None

    Example:
        Args:
            bucket = 'MyBucket'
            schema = 'MySchema'
            table = 'MyTable'
            filepath = 'path/to/data/apple=abcd/banana=1234/abcd1234.parquet'
            session_helper = some_configured_session
    '''
    partitions = _get_partitions_for_spectrum(filepath)
    formatted_partitions = _format_partition_strings_for_sql(partitions)
    path_to_data = _get_partition_location(filepath)

    with session_helper.db_session_scope() as scope:
        partitions_query = f"ALTER TABLE {schema}.{table} \
            ADD IF NOT EXISTS PARTITION ({', '.join(formatted_partitions)}) \
            LOCATION 's3://{bucket}/{path_to_data}';"
        logger.info(f'Running query to create: {partitions_query}')
        scope.execute(partitions_query)



import ast
import boto3
from collections import OrderedDict
import datetime
from distutils.util import strtobool
import logging
import operator
from typing import Dict, List, Any
import multiprocessing as mp
from multiprocessing import get_context
import s3fs
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
from .s3_naming_helper import S3NamingHelper

# Filter
# class Filter(NamedTuple):
#    partition: str
#    comparison: str
#    values: List[any]

Filter = {
    "partition": str,
    "comparison": str,
    "values": List[any]
}

OPS = {
    "==": operator.eq,
    "!=": operator.ne,
    ">=": operator.ge,
    "<=": operator.le,
    ">": operator.gt,
    "<": operator.lt
}

NON_NUM_TYPES = [
    "string",
    "category",
    "bool",
    "boolean"
]


logger = logging.getLogger(__name__)


def get_all_partition_values(bucket: str, key: str, partition: str) -> iter:
    """ Returns all values, correctly typed, for a given partition IN NO ORDER

    Args:
        bucket (str): S3 Bucket name
        key (str): S3 key that leads to the desired dataset
        partition (str): The partition in the dataset to parse the values out of

    Returns:
        An iterable of all the partition values
    """

    all_files = get_all_files_list(bucket, key)

    if not all_files:
        return []

    partition_dtype = _get_partitions_and_types(
        first_file_key=all_files[0], bucket=bucket)[partition]
    partition_values = _parse_partitions_and_values(all_files, key=key)[
        partition]
    return [convert_type(val, partition_dtype) for val in partition_values]


def get_diff_partition_values(bucket: str, key: str, partition: str, values_to_diff: iter, reverse: bool = False) -> iter:
    """ Returns all the partition values in the dataset at the bucket/key
    that are not in values_to_diff

    Args:
        bucket (str): S3 Bucket name
        key (str): S3 key that leads to the desired dataset
        partition (str): The partition in the dataset to parse the values out of
        values_to_diff (iter): The iterable set of values to compare the partition values to
        reverse (bool, Optional): Determines if the operation should be inversed,
            if True it will look for the values in values_to_diff that are not
            in the partition values (basically backwards). Defaults to False

    Returns:
        An iterable of all the partition values that are not in values_to_diff
            and vice-versa if reverse is True
    """
    all_files = get_all_files_list(bucket, key)

    if not all_files:
        if reverse:
            # Keeping  ->set->list fuctionality to have consistent lack of duplicates
            diff = list(set(values_to_diff))
            return diff
        else:
            return []

    partition_dtype = _get_partitions_and_types(
        first_file_key=all_files[0], bucket=bucket)[partition]
    partition_values = _parse_partitions_and_values(all_files, key=key)[
        partition]

    partition_set = set(partition_values)
    values_to_diff_set = set([str(val) for val in values_to_diff])

    if not values_to_diff:
        if reverse:
            return []
        else:
            return [convert_type(val, partition_dtype) for val in partition_set]

    if reverse:
        diff = values_to_diff_set - partition_set
    else:
        diff = partition_set - values_to_diff_set
    return [convert_type(val, partition_dtype) for val in diff]


def get_max_partition_value(bucket: str, key: str, partition: str) -> any:
    """ Returns the max partition value in the dataset at the bucket/key
    NOTE: This functionality cannot occur on non-numerical datatypes!

    Args:
        bucket (str): S3 Bucket name
        key (str): S3 key that leads to the desired dataset
        partition (str): The partition in the dataset to parse the values out of

    Returns:
        The max partition value, of the matching partition datatype
    """
    S3NamingHelper().validate_bucket_name(bucket_name=bucket)

    all_files = get_all_files_list(bucket=bucket, key=key)

    if not all_files:
        return None

    partition_dtype = _get_partitions_and_types(
        first_file_key=all_files[0], bucket=bucket)[partition]
    partition_values = _parse_partitions_and_values(
        file_paths=all_files, key=key)[partition]

    if partition_dtype in NON_NUM_TYPES:
        raise ValueError(
            f"Max cannot be used on partition types of {partition_dtype}")

    return max([convert_type(val, partition_dtype) for val in partition_values])


def fetch(bucket: str, key: str, filters: List[type(Filter)] = {}, parallel: bool = True, accept_not_s3parq: bool = True) -> pd.DataFrame:
    """ S3 Parquet to Dataframe Fetcher.
    This function handles the portion of work that will return a concatenated 
    dataframe based on the partition filters of the specified dataset.

    Args:
        bucket (str): S3 Bucket name
        key (str): S3 key that leads to the desired dataset
        filters (List(Filter), Optional): filters to be applied to the Dataset Partitions
            Filters have the fields:

                - partition (str):
                    The published partition to be applied to
                - comparison (str):
                    Comparison function - one of: [ == , != , > , < , >= , <= ]
                - values (List(any)): 
                    Values to compare to - must match partition data type
                    May not use multiple values with the '<', '>' comparisons
        parallel (bool, Optional):
            Determines if multiprocessing should be used, defaults to True
            Whether this is more or less efficient is dataset dependent,
                smaller datasets run much quicker without parallel
        accept_not_s3parq (bool, Optional):
            Determines whether it will fetch dataframes that have not been published
                with S3Parq and are thus lacking metadata. This does not support
                partitioned parquet.

    Returns:
        A pandas dataframe of the filtered results from S3
    """

    _validate_filter_rules(filters)
    S3NamingHelper().validate_bucket_name(bucket)

    all_files = get_all_files_list(bucket, key)

    if not all_files:
        logger.debug(f"No files present under : {key} :, returning empty DataFrame")
        return pd.DataFrame()

    partition_metadata = _get_partitions_and_types(all_files[0], bucket)

    if partition_metadata is None:
        if accept_not_s3parq:
            logger.info("Parquet files do not have S3Parq metadata, fetching anyways.")
            return _get_filtered_data(bucket=bucket, paths=all_files, partition_metadata={},
                              parallel=parallel)
        else:
            raise MissingS3ParqMetadata("Parquet files are missing s3parq metadata, enable 'accept_not_s3parq' if you'd like this to pass.")

    _validate_matching_filter_data_type(partition_metadata, filters)

    # strip out the filenames (which we need)
    partition_values = _parse_partitions_and_values(all_files, key)

    typed_values = _get_partition_value_data_types(
        partition_values, partition_metadata)

    # filtered_paths is a list of the S3 prefixes from which we want to load data
    filtered_paths = _get_filtered_key_list(typed_values, filters, key)

    files_to_load = []

    for file in all_files:
        for prefix in filtered_paths:
            if file.startswith(prefix):
                files_to_load.append(file)

    # if there is no data matching the filters, return an empty DataFrame
    # with correct headers and type
    if len(files_to_load) < 1:
        logger.debug(f"Dataset filters left no matching values, returning empty dataframe with matching headers")
        sacrifical_files = [all_files[0]]
        sacrifical_frame = _get_filtered_data(
            bucket=bucket, paths=sacrifical_files, partition_metadata=partition_metadata, parallel=parallel)
        return sacrifical_frame.head(0)

    return _get_filtered_data(bucket=bucket, paths=files_to_load, partition_metadata=partition_metadata,
                              parallel=parallel)


def fetch_diff(input_bucket: str, input_key: str, comparison_bucket: str, comparison_key: str, partition: str, reverse: bool = False, parallel: bool = True) -> pd.DataFrame:
    """ Returns a dataframe of whats in the input dataset but not the comparison 
    dataset by the specified partition.

    Args:
        input_bucket (str): The bucket of the dataset to start from
        input_key (str): The key to the dataset to start from
        comparison_bucket (str): The bucket of the dataset to compare against
        comparison_key (str): The key to the dataset to compare against
        partition (str): The partition in the dataset to compare the values of
        reverse (bool, Optional): Determines if the operation should be inversed,
            if True it will look for the values in comparison that are not 
            in the input (basically backwards). Defaults to False
        parallel (bool, Optional):
            Determines if multiprocessing should be used, defaults to True.
            Whether this is more or less efficient is dataset dependent,
            smaller datasets run much quicker without parallel

    Returns:
        A dataframe of all values in the input but not the comparison,
            if reverse=True then vice-versa
    """
    S3NamingHelper().validate_bucket_name(input_bucket)
    S3NamingHelper().validate_bucket_name(comparison_bucket)

    comparison_values = get_all_partition_values(
        bucket=comparison_bucket, key=comparison_key, partition=partition)

    diff_values = get_diff_partition_values(
        bucket=input_bucket,
        key=input_key,
        partition=partition,
        values_to_diff=comparison_values,
        reverse=reverse
    )

    filters = [{
        "partition": partition,
        "comparison": "==",
        "values": diff_values
    }]

    logger.debug(f"Fetching difference, looking for values : {diff_values} : under partition : {partition} ")

    if reverse:
        return fetch(bucket=comparison_bucket, key=comparison_key, filters=filters, parallel=parallel)
    else:
        return fetch(bucket=input_bucket, key=input_key, filters=filters, parallel=parallel)


def convert_type(val: Any, dtype: str) -> Any:
    """ Converts the given value to the given datatype

    Args:
        val (Any): The value to convert
        dtype (str): The type to attempt to convert to

    Returns:
        The value parsed into the new dtype
    """
    if dtype == 'string' or dtype == 'str':
        return str(val)
    elif dtype == 'integer' or dtype == 'int':
        return int(val)
    elif dtype == 'float':
        return float(val)
    elif dtype == 'datetime':
        return datetime.datetime.strptime(
            val, '%Y-%m-%d %H:%M:%S')
    elif dtype == 'category':
        return pd.Category(val)
    elif dtype == 'bool' or dtype == 'boolean':
        return bool(strtobool(val))


def dtype_to_pandas_dtype(dtype: str) -> str:
    """ Matches the given dtype to its pandas equivalent, if one exists

    Args:
        dtype (str): The type to attempt to match

    Returns:
        The pandas version of the dtype if one exists, otherwise the original
    """
    if dtype == "integer":
        dtype = "int"
    elif dtype == "string":
        dtype = "str"
    elif dtype == "boolean":
        dtype = "bool"

    return dtype


def get_all_files_list(bucket: str, key: str) -> list:
    """ Get a list of all files to get all partitions values.
    Necesarry to catch all partition values for non-filtered partiions.
    NOTE: This paginates across as many as are matching keys in the bucket;
        be mindful of any costs in extra large buckets without a specific key.

    Args:
        bucket (str): S3 bucket to search in
        key (str): S3 key to the dataset to check

    Returns:
        A list of the keys of all objects in the bucket/key that end with .parquet
    """
    objects_in_bucket = []
    s3_client = boto3.client('s3')
    paginator = s3_client.get_paginator('list_objects')
    operation_parameters = {'Bucket': bucket,
                            'Prefix': key}
    page_iterator = paginator.paginate(**operation_parameters)
    for page in page_iterator:
        if not "Contents" in page.keys():
            break

        for item in page['Contents']:
            if item['Key'].endswith('.parquet'):
                objects_in_bucket.append(item['Key'])

    return objects_in_bucket


def _get_partitions_and_types(first_file_key: str, bucket: str) -> dict:
    """ Fetch a list of all the partitions actually there and their 
    datatypes. List may be different than passed list if not being used
    for filtering on.
    NOTE: This is pulled from the metadata. It is assumed that this package
        is being used with its own publish method.

    Args:
        first_file_key (str): The key to the first file in the dataset, to get
            the metadata off of
        bucket (str): The S3 bucket to run in

    Returns:
        The parsed metadata from heading the first file
    TODO
    """
    s3_client = boto3.client('s3')

    first_file = s3_client.head_object(
        Bucket=bucket,
        Key=first_file_key
    )

    metadata = first_file['Metadata']
    partition_metadata_raw = metadata.get("partition_data_types", None)

    if partition_metadata_raw is None:
        return None

    partition_metadata = ast.literal_eval(partition_metadata_raw)

    return partition_metadata


def _parse_partitions_and_values(file_paths: List[str], key: str) -> dict:
    """ Parses the string keys into a usable dict of the parts and values

    Args:
        file_paths (List[str]): The list of all files to parse out
        key (str): S3 key to the root of the dataset of the files

    Returns:
        A dictionary of all partitions with their values
    """
    # TODO: find more neat/efficient way to do this
    parts = OrderedDict()
    key_len = len(key)
    for file_path in file_paths:
        # Delete key, split the parts out, delete the file name
        file_path = file_path[key_len:]
        unparsed_parts = file_path.split("/")
        del unparsed_parts[-1]

        for part in unparsed_parts:
            if "=" in part:
                key, value = part.split("=")
                if key not in parts:
                    parts.update({key: set([value])})
                else:
                    parts[key].add(value)

    return parts


def _get_partition_value_data_types(parsed_parts: dict, part_types: dict) -> dict:
    """ Uses the partitions with their known types to parse them out

    Args:
        parsed_parts (dict): A dictionary of all partitions with their values
        part_types (dict): A dictionary of all partitions to their datatypes

    Returns:
        A dictionary of all partitions with their values parsed into the correct datatype
    """
    for part, values in parsed_parts.items():
        part_type = part_types[part]
        if (part_type == 'string') or (part_type == 'category'):
            continue
        elif (part_type == 'int') or (part_type == 'integer'):
            parsed_parts[part] = set(map(int, values))
        elif part_type == 'float':
            parsed_parts[part] = set(map(float, values))
        elif part_type == 'datetime':
            parsed_parts[part] = set(
                map(lambda s: datetime.datetime.strptime(s, "%Y-%m-%d %H:%M:%S"), values))
        elif (part_type == 'bool') or (part_type == 'boolean'):
            parsed_parts[part] = set(map(bool, values))
        else:
            logger.debug(f"Unknown partition type : {part_type} :, leaving as a string")

    return parsed_parts


def _get_filtered_key_list(typed_parts: dict, filters: List[type(Filter)], key: str) -> List[str]:
    """ Create list of all "paths" to files after the filtered partitions
    are set ie all non-matching partitions are excluded.

    Args:
        typed_parts (dict): A dictionary of all partitions with their values
        filters (List[type(Filter)]): A dictionary of all partitions to their datatypes
        key (str): S3 key to the base root of the dataset

    Returns:
        A list of object keys that have been filtered by partition values
    """
    filter_keys = []
    matched_parts = OrderedDict()
    matched_parts.keys = typed_parts.keys()
    matched_part_vals = set()

    for part, part_values in typed_parts.items():
        matched_part_vals.clear()
        fil = next((f for f in filters if f['partition'] == part), False)
        if fil:
            comparison = OPS[fil['comparison']]
            for v in fil['values']:
                for x in part_values:
                    if comparison(x, v):
                        matched_part_vals.add(x)
            matched_parts[part] = matched_part_vals.copy()
        else:
            matched_parts[part] = part_values

    def construct_paths(matched_parts, previous_fil_keys: List[str]) -> None:
        if len(matched_parts) > 0:
            part = matched_parts.popitem(last=False)
            new_filter_keys = list()
            for value in part[1]:
                mapped_keys = list(map(
                    (lambda x: str(x) +
                     str(part[0]) + "=" + str(value) + "/"),
                    previous_fil_keys
                ))
                new_filter_keys = new_filter_keys + mapped_keys

            construct_paths(matched_parts, new_filter_keys)
        else:
            filter_keys.append(previous_fil_keys)

    construct_paths(matched_parts, [f"{key}/"])
    # TODO: fix the below mess with random array
    return filter_keys[0]


def _get_filtered_data(bucket: str, paths: List[str], partition_metadata: dict, parallel: bool = True) -> pd.DataFrame:
    """ Gets the data based on the filtered object key list. Concatenates all 
    the separate parquet files.

    Args:
        bucket (str): S3 bucket to fetch from
        paths (List[str]): A list of all the object keys of the parquet files
        partition_metadata (dict): A dictionary of all partitions to their datatypes
        parallel (bool, Optional):
            Determines if multiprocessing should be used, defaults to True.
            Whether this is more or less efficient is dataset dependent,
            smaller datasets run much quicker without parallel

    Returns:
        The dataframe of all the parquet files from the given keys
    """

    temp_frames = []

    def append_to_temp(frame: pd.DataFrame):
        temp_frames.append(frame)

    if parallel:
        with get_context("spawn").Pool() as pool:
            for path in paths:
                logger.debug(f"[Parallel] Fetching parquet file to append from : {path} ")
                append_to_temp(
                    pool.apply_async(_s3_parquet_to_dataframe, args=(bucket, path, partition_metadata)).get())
            pool.close()
            pool.join()
    else:
        for path in paths:
            logger.debug(f"Fetching parquet file to append from : {path} ")
            append_to_temp(_s3_parquet_to_dataframe(
                bucket, path, partition_metadata))

    return pd.concat(temp_frames, ignore_index=True)


def _s3_parquet_to_dataframe(bucket: str, key: str, partition_metadata: dict) -> pd.DataFrame:
    """ Grab a parquet file from s3 and convert to pandas df, add it to the destination

    Args:
        bucket (str): S3 bucket to fetch from
        key (key): The full object path of the parquet file to read
        partition_metadata (dict): A dictionary of all partitions to their datatypes

    Returns:
        Dataframe from the key parquet, with partitions repopulated
    """
    s3 = s3fs.S3FileSystem()
    uri = f"{bucket}/{key}"
    table = pq.ParquetDataset(uri, filesystem=s3)
    frame = table.read_pandas().to_pandas()

    if partition_metadata != {}:
        partitions = _repopulate_partitions(uri, partition_metadata)
        for k, v in partitions.items():
            frame[k] = v

    return frame


def _repopulate_partitions(partition_string: str, partition_metadata: dict) -> tuple:
    """ For each val in the partition string creates a list that can be added back into the dataframe

    Args:
        partition_string (str): S3 bucket to fetch from
        partition_metadata (dict): A dictionary of all partitions to their datatypes

    Returns:
        Dataframe from the key parquet, with partitions repopulated
    """
    raw = partition_string.split('/')
    partitions = {}
    for string in raw:
        if '=' in string:
            k, v = string.split('=')
            partitions[k] = v

    if partition_metadata:
        for key, val in partitions.items():
            partitions[key] = convert_type(val, partition_metadata[key])
    
    return partitions


def _validate_filter_rules(filters: List[type(Filter)]) -> None:
    """ Validate that the filters are the correct format and follow basic
    comparison rules, otherwise throw a ValueError

    Args:
        filters (List[type(Filter)]): List of filters to validate

    Returns:
        None
    """

    single_value_comparisons = [
        ">",
        "<",
        "<=",
        ">="
    ]
    for f in filters:
        if not all(key in f for key in ("partition", "comparison", "values")):
            raise ValueError(
                "Filters require partition, comparison, and values.")
        elif f["comparison"] not in OPS:
            raise ValueError(
                f"Comparison {f['comparison']} is not supported.")
        elif (f["comparison"] in single_value_comparisons) & (len(f["values"]) != 1):
            raise ValueError(
                f"Comparison {f['comparison']} can only be used with one filter value.")


def _validate_matching_filter_data_type(part_types, filters: List[type(Filter)]) -> None:
    """ Validate that the filters passed are matching to the partitions'
    listed datatypes, otherwise throw a ValueError. 
    This includes validating comparisons too.

    Args:
        part_types (dict): A dictionary of all partitions to their datatypes
        filters (List[type(Filter)]): List of filters to validate

    Returns:
        None
    """
    num_comparisons = [
        ">",
        "<",
        "<=",
        ">="
    ]

    for f in filters:
        try:
            fil_part = part_types[f["partition"]]
        except:
            raise ValueError("Filter does not have a matching partition.")

        if (f["comparison"] in num_comparisons):
            if fil_part in NON_NUM_TYPES:
                raise ValueError(
                    f"Comparison {f['comparison']} cannot be used on partition types of {fil_part}")


# Custom error for catching
class MissingS3ParqMetadata(Exception):
    """ This class is meant to be an Exception thrown only when parquet files
    being fetched are not S3Parq compatible and the option to fetch anyways is
    disabled.
    """
    pass



import boto3
import sqlalchemy
from sqlalchemy.orm import sessionmaker
from contextlib import contextmanager
import urllib
import urllib.request
import logging


class SessionHelper:
    """
    Helper class for establishing a connection with a Redshift database.  It assumes your IAM
    role has been configured to access the cluster _and_ has privileges on the database.

    Args:
        region (str): AWS region the session should be using for boto
        cluster_id (str): AWS Redshift cluser_id to connect to
        host (str): Host for Redshift connection
        port (str): port for Redshift connection
        db_name (str): Database name to connect to in Redshift
        ec2_user (str): User to use for ec2, can be blank if using iam of local

    Example:
        SH = SessionHelper('us-east-1', 'some-cluster-1', 'some-host','5439','some-database')
        SH.configure_session_helper()
    """

    def __init__(self, region: str, cluster_id: str, host: str, port: str, db_name: str, ec2_user: str):
        self.region = region
        self.cluster_id = cluster_id
        self.host = host
        self.port = port
        self.db_name = db_name
        self.ec2_user = ec2_user
        self.is_ec2_flag = self._is_ec2()

    def _is_ec2(self):
        """ Determines if the session is running on an ec2 server.
        Uses urllib to determine this by attempting to find the instance id;
        if there is one then its on ec2, otherwise the command will hang and
        timeout.
        """
        try:
            # Timeout set to 2 seconds to avoid prolonged hang if the url cannot be reached
            self.instance_id = urllib.request.urlopen(
                'http://169.254.169.254/latest/meta-data/instance-id', timeout=2).read().decode()
            return True
        except urllib.error.URLError:
            return False

    def set_boto_session(self) -> None:
        """ Adds the boto3 session to the class, with the region set here """
        self.boto_session = boto3.Session(region_name=self.region)

    def set_iam_user(self) -> None:
        """ Adds the iam user to the SessionHelper, has a switch for ec2 or local """
        if self.is_ec2_flag:
            # On ec2, the cluster user is set from the redshift configuration dictionary
            self.iam_user = self.ec2_user
        else:
            iam_client = self.boto_session.client('iam')
            iam_user = iam_client.get_user()
            self.iam_user = iam_user['User']['UserName']

    def set_aws_credentials(self, session: boto3.Session) -> None:
        """ Sets the aws_credentials based on what's in the boto session """
        credentials = session.get_credentials()
        self.aws_credentials = credentials.get_frozen_credentials()

    def make_db_session(self, **kwargs) -> None:
        """ Creates the sqlalchemy Session to Redshift based on the class params
        NOTE: this session is set to autocommit

        Args:
            **kwargs : Arbitrary length keyword params
                Must include:
                    user (str): The user to connect to Redshift under
                    pwd (str): The password to connect to Redshift with

        Returns:
            None
        """
        user, pwd = kwargs['user'], kwargs['pwd']
        self.engine = sqlalchemy.create_engine(
            f'postgresql://{user}:{pwd}@{self.host}:{self.port}/{self.db_name}', isolation_level="AUTOCOMMIT")
        self.Session = sessionmaker(bind=self.engine)

    def get_redshift_credentials(self) -> dict:
        """ Uses the boto session client to get temporary credentials to
        connect to Redshift with.
        If this is in ec2, it will just use the local access stuff, otherwise
        it will use the boto session's credentials.

        Returns:
            This returns a dict of the cluster_credentials per boto's Redshift
            Format is:
                DbUser (str): Database user for Redshift
                DbPassword (str): Password for the database user for Redshift
                Expiration (datetime): The date and time when the password expires
        """
        if self.is_ec2_flag:
            client = self.boto_session.client(
                'redshift', region_name=self.region)

            temp_redshift_credentials = client.get_cluster_credentials(
                DbUser=self.iam_user,
                ClusterIdentifier=self.cluster_id,
                AutoCreate=True,
            )
        else:
            client = self.boto_session.client(
                'redshift',
                region_name=self.region,
                aws_access_key_id=self.aws_credentials.access_key,
                aws_secret_access_key=self.aws_credentials.secret_key
            )
            temp_redshift_credentials = client.get_cluster_credentials(
                DbUser=self.iam_user,
                ClusterIdentifier=self.cluster_id,
                AutoCreate=True,
            )
        return temp_redshift_credentials

    def parse_temp_redshift_credentials(self, rs_creds: dict) -> tuple:
        """ Return the user credentials as a tuple, with the username html safe
        by changing any spaces the +'s

        Args:
            rs_creds (dict): The Redshift credentials returned by boto
                Format is:
                    DbUser (str): Database user for Redshift
                    DbPassword (str): Password for the database user for Redshift
                    Expiration (datetime): The date and time when the password expires

        Returns:
            Tuple of creds, format of:
                DbUser (str): Database user for Redshift, html escaped
                DbPassword (str): Password for the database user for Redshift
        """
        username = rs_creds['DbUser']
        username = urllib.parse.quote_plus(username)
        pwd = rs_creds['DbPassword']
        return username, pwd

    def configure_session_helper(self) -> None:
        """ Runs all of the configuration steps for SessionHelper in one go.
        Steps are:
            1. Creates boto session with the initialized region
            2. Sets the iam user
            3. Sets the full aws credentials from the boto session
            4. Gets temporary Redshift credentials
            5. Creates the sqlalchemy database session for use based on above
        """
        try:
            self.set_boto_session()
            self.set_iam_user()
            self.set_aws_credentials(self.boto_session)
            temp_creds = self.get_redshift_credentials()
            user, pwd = self.parse_temp_redshift_credentials(temp_creds)
            self.make_db_session(user=user, pwd=pwd)
        except Exception as e:
            logging.error('Failed to set up the SessionHelper')
            logging.error(e)
            raise

    @contextmanager
    def db_session_scope(self) -> sessionmaker:
        """ Wraps the sqlalchemy session so that it can used a try-except-finally
        to ensure there are rollbacks and that the session is always closed.

        Yields:
            The sqlalchemy database session, configured by class attributes

        Excepts:
            Exception: This triggers a session rollback, after which it is
                raised immediately
        """
        session = self.Session()
        try:
            yield session
            session.commit()
        except Exception as e:
            session.rollback()
            raise e
        finally:
            session.close()



from setuptools import setup, find_packages
from os import path

package_name = "s3parq"
package_version = "2.1.14"
description = "Write and read/query s3 parquet data using Athena/Spectrum/Hive style partitioning."
cur_directory = path.abspath(path.dirname(__file__))
with open(path.join(cur_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name=package_name,
    version=package_version,
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Integrichain Innovation Team",
    author_email="engineering@integrichain.com",
    url="https://github.com/IntegriChain1/s3parq",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7"
        ],
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    install_requires=[
            "pandas>=0.24.2, <1",
            "pyarrow>=0.13.0",
            "boto3>=1.9",
            "s3fs>=0.2",
            "psycopg2==2.8.3",
            "SQLAlchemy>=1.3"
    ]
)



__version__ = "2.1.14"

from s3parq.fetch_parq import fetch
from s3parq.fetch_parq import fetch_diff
from s3parq.fetch_parq import get_all_files_list
from s3parq.fetch_parq import get_max_partition_value
from s3parq.publish_parq import publish
from s3parq.session_helper import SessionHelper
